#!/usr/bin/env python
#
# The software in this package is distributed under the GNU General
# Public License version 2 (with a special exception described below).
#
# A copy of GNU General Public License (GPL) is included in this distribution,
# in the file COPYING.GPL.
#
# As a special exception, if other files instantiate templates or use macros
# or inline functions from this file, or you compile this file and link it
# with other works to produce a work based on this file, this file
# does not by itself cause the resulting work to be covered
# by the GNU General Public License.
#
# However the source code for this file must still be made available
# in accordance with section (3) of the GNU General Public License.
#
# This exception does not invalidate any other reasons why a work based
# on this file might be covered by the GNU General Public License.
#
# Copyright (c) 2016-2018 Intra2net AG <info@intra2net.com>

"""

summary
-------------------------------------------------------------------------------
Represent CNF_VARs as recursive structures.

Copyright: 2014-2017 Intra2net AG
License:   GPLv2+


contents
-------------------------------------------------------------------------------

This module provides read and write functionality for the Intra2net *CNF*
format. Two different syntaxes are available: classical *CNF* and a JSON
representation. Both versions are commonly understood by Intra2net software.

On the command line, raw CNF is accepted if the option ``-`` is given: ::

    $ get_cnf routing 2 |python3 cnfvar.py - <<ENOUGH
    1 ROUTING,2: "192.168.55.0"
    2    (1) ROUTING_COMMENT,0: ""
    3    (1) ROUTING_DNS_RELAYING_ALLOWED,0: "1"
    4    (1) ROUTING_EMAIL_RELAYING_ALLOWED,0: "1"
    5    (1) ROUTING_FIREWALL_RULESET_REF,0: "9"
    6    (1) ROUTING_GATEWAY,0: "10.0.254.1"
    7    (1) ROUTING_NAT_INTO,0: "0"
    8    (1) ROUTING_NETMASK,0: "255.255.255.0"
    9    (1) ROUTING_PROXY_PROFILE_REF,0: "2"
    ENOUGH

    1 ROUTING,2: "192.168.55.0"
    2   (1) ROUTING_COMMENT,0: ""
    3   (1) ROUTING_DNS_RELAYING_ALLOWED,0: "1"
    4   (1) ROUTING_EMAIL_RELAYING_ALLOWED,0: "1"
    5   (1) ROUTING_FIREWALL_RULESET_REF,0: "9"
    6   (1) ROUTING_GATEWAY,0: "10.0.254.1"
    7   (1) ROUTING_NAT_INTO,0: "0"
    8   (1) ROUTING_NETMASK,0: "255.255.255.0"
    9   (1) ROUTING_PROXY_PROFILE_REF,0: "2"

The input takes one round-trip through the parsers and will error out on
problematic lines. Thus, ``cnfvar.py`` can be used to syntax-check CNF data.

Note that line numbers may be arbitrarily reassigned in the process. Of course,
parent references and the relative ordering of lines will be preserved in this
case.

.. todo::
    Decide on some facility for automatic fixup of line number values. The
    internal representation is recursive so line numbers are not needed to
    establish a variable hierarchy. They might as well be omitted from the json
    input and only added when writing cnf. For the time being a lack of the
    "number" field is interpreted as an error. Though it should at least
    optionally be possible to omit the numbers entirely and have a function add
    them after the fact. This might as well be added to :py:func:`is_cnf`
    though it would be counterintuitive to have a predicate mutate its
    argument. So maybe it could return a second argument to indicate a valid
    structure that needs fixup or something like that.

.. note::
    The variable values of get_cnf seems to be encoded in latin1, and set_cnf
    seems to assume latin1-encoded values (not var names). Function
    :py:func:`read_cnf` converts this to unicode and functions
    :py:func:`dump_cnf_string`, :py:func:`print_cnf`, and :py:func:`write_cnf`
    convert unicode back to latin1.


notes on Python 3 conversion
-------------------------------------------------------------------------------

Since the original *CNF* format assumes latin-1 encoded data pretty much
exclusively, we preserve the original encoding while parsing the file.
When assembling the data structures returned to the user, values are then
converted to strings so they can be used naturally at the Python end.

implementation
-------------------------------------------------------------------------------
"""

import functools
import sys
import json
import re
import io

#
#                                   helpers
#

# Sadly, the Intranator is still stuck with one leg in the 90s.
#


def to_latin1(s):
    """ take given unicode value and convert it to a latin1-encoded string """
    return s.encode("latin-1")


def from_latin1(s):
    """ take given latin1-encoded string value and convert it to unicode """
    return s.decode("latin-1")

#
#                                  traversal
#


class InvalidCNF(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "Malformed CNF_VAR: \"%s\"" % self.msg


def walk_cnf(cnf, nested, fun, acc):
    """
    :type cnf: cnf list
    :type nested: bool
    :type fun: 'a -> bool -> (cnf stuff) -> 'a
    :type acc: 'a
    :rtype: 'a
    """
    for var in cnf:
        acc = fun(acc,
                  nested,
                  comment=var.get("comment", None),
                  data=var.get("data", None),
                  instance=var.get("instance", None),
                  number=var.get("number", None),
                  parent=var.get("parent", None),
                  varname=var.get("varname", None))
        children = var.get("children", None)
        if children is not None:
            acc = walk_cnf(children, True, fun, acc)
    return acc


#
#                                 validation
#

def is_string(s):
    return isinstance(s, str)


def is_valid(acc,
             nested,
             comment,
             data,
             instance,
             number,
             parent,
             varname):
    if varname is None:
        raise InvalidCNF("CNF_VAR lacks a name.")
    elif not is_string(varname):
        raise InvalidCNF("Varname field of CNF_VAR \"%s\" is not a string."
                         % varname)
    elif varname == "":
        raise InvalidCNF("Varname field of CNF_VAR is the empty string.")

    if comment is not None:
        if not is_string(comment):
            raise InvalidCNF("Comment field of CNF_VAR \"%s\" is not a string."
                             % varname)

    if data is None:
        raise InvalidCNF("Data field of CNF_VAR \"%s\" is empty."
                         % varname)
    elif not is_string(data):
        raise InvalidCNF("Data field of CNF_VAR \"%s\" is not a string."
                         % varname)

    if instance is None:
        raise InvalidCNF("Instance field of CNF_VAR \"%s\" is empty."
                         % varname)
    elif not isinstance(instance, int):
        raise InvalidCNF("Instance field of CNF_VAR \"%s\" is not an integer."
                         % varname)

    if number is None:
        raise InvalidCNF("Number field of CNF_VAR \"%s\" is empty."
                         % varname)
    elif not isinstance(number, int):
        raise InvalidCNF("Number field of CNF_VAR \"%s\" is not an integer."
                         % varname)
    elif number < 1:
        raise InvalidCNF("Number field of CNF_VAR \"%s\" must be positive, not %d."
                         % (varname, number))
    else:
        other = acc.get(number, None)
        if other is not None:  # already in use
            raise InvalidCNF("Number field of CNF_VAR \"%s\" already used by variable %s."
                             % (varname, other))
        acc[number] = varname

    if nested is True:
        if parent is None:
            raise InvalidCNF("Parent field of nested CNF_VAR \"%s\" is empty."
                             % varname)
        elif not isinstance(parent, int):
            raise InvalidCNF("Parent field of CNF_VAR \"%s\" is not an integer."
                             % varname)
    else:
        if parent is not None:
            raise InvalidCNF("Flat CNF_VAR \"%s\" has nonsensical parent field \"%s\"."
                             % (varname, parent))
    return acc


def is_cnf(root):
    """
    is_cnf -- Predicate testing "CNF_VAR-ness". Folds the :py:func:`is_valid`
    predicate over the argument which can be either a well-formed CNF
    dictionary or a list of CNF_VARs.

    :type root: cnfvar or cnf list
    :rtype: bool

    Not that if it returns at all, ``is_cnf()`` returns ``True``. Any non
    well-formed member of the argument will cause the predicate to bail out
    with an exception during traversal.
    """
    cnf = cnf_root (root)
    if cnf is None:
        raise InvalidCNF (root)
    return walk_cnf(cnf, False, is_valid, {}) is not None


def count_vars (root):
    """
    Traverse the cnf structure recursively, counting VAR objects (CNF lines).
    """
    cnf = cnf_root (root)
    if cnf is None:
        raise InvalidCNF (root)
    return walk_cnf (cnf, True, lambda n, _nested, **_kwa: n + 1, 0)

#
#                               deserialization
#

# the easy part: JSON reader for get_cnf -j

def make_varname_lowercase(cnfvar):
    """
    Custom hook for json decoder: convert variable name to lowercase.

    Since variable names are case insensitive, :py:func:`read_cnf` converts
    them all to lower case. Downstream users of :py:func:`read_cnf_json` (e.g.
    :py:class:`simple_cnf.SimpleCnf`) rely on lowercase variable names.

    :param dict cnfvar: JSON "object" converted into a dict
    :returns: same as input but if field `varname` is present, its value is
              converted to lower case
    :rtype: dict with str keys
    """
    try:
        cnfvar['varname'] = cnfvar['varname'].lower()
    except KeyError:      # there is no "varname" field
        pass
    except AttributeError:   # cnfvar['varname'] is not a string
        pass
    return cnfvar


def read_cnf_json(cnfdata):
    """
    Read json data from cnf data bytes.

    :param bytes cnfdata: config data
    :return: the parsed json data
    :rtype: str

    .. note:: The JSON module does not decode data for all versions
        of Python 3 so we handle the decoding ourselves.
    """
    if isinstance (cnfdata, bytes) is True:
        cnfdata = from_latin1 (cnfdata)
    cnf_json = json.loads(cnfdata, object_hook=make_varname_lowercase)
    if is_cnf(cnf_json) is False:
        raise TypeError("Invalid CNF_VAR.")
    return cnf_json

# the moderately more complicated part: CNF reader


def advance(cns):
    current, next, stream = cns
    if next is None:  # reached end of stream
        return None
    current = next

    try:
        next = stream.pop()
        next = next.strip()
    except IndexError:
        next = None

    if current == "":  # skip blank lines
        return advance((current, next, stream))
    return (current, next, stream)


def prepare(raw):
    """
    :type raw: bytes
    :rtype: (str * str option * str list) option
    """
    lines = raw.splitlines()
    lines.reverse()
    try:
        first = lines.pop()
    except IndexError:
        return None

    try:
        second = lines.pop()
    except IndexError:
        second = None

    first = first.strip()
    if first == b"":
        return advance((first, second, lines))

    return (first, second, lines)


def peek(cns):
    _, next, _ = cns
    return next


def get(cns):
    current, _, _ = cns
    return current


class MalformedCNF(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "Malformed CNF file: \"%s\"" % self.msg

grab_parent_pattern = re.compile(b"""
                                    ^            # match from start
                                    \d+          # line number
                                    \s+          # spaces
                                    \((\d+)\)    # parent
                                 """,
                                 re.VERBOSE)


def get_parent(line):
    match = re.match(grab_parent_pattern, line)
    if match is None:  # -> no parent
        return None
    return int(match.groups()[0])


def marshal_in_number   (number):   return int (number)

def marshal_in_parent   (parent):   return int (parent)

def marshal_in_varname  (varname):  return from_latin1 (varname).lower ()

def marshal_in_instance (instance): return int (instance)

def marshal_in_data     (data):     return from_latin1 (data) if data is not None else ""

def marshal_in_comment  (comment):  return comment and from_latin1(comment[1:].strip()) or None

base_line_pattern = re.compile(b"""
                                    ^                    # match from start
                                    \s*                  # optional spaces
                                    (\d+)                # line number
                                    \s+                  # spaces
                                    ([A-Z][A-Z0-9_]*)    # varname
                                    \s*                  # optional spaces
                                    ,                    # delimiter
                                    \s*                  # optional spaces
                                    (-1|\d+)             # instance
                                    \s*                  # optional spaces
                                    :                    # delimiter
                                    \s*                  # optional spaces
                                    \"(                  # quoted string (data)
                                       (?:  \\\"         #  (of escaped dquote
                                           |[^\"])*      #   or anything not a
                                      )\"                #   literal quote)
                                    \s*                  # optional spaces
                                    (                    # bgroup
                                     \#                  # comment leader
                                     \s*                 # optional spaces
                                     .*                  # string (comment)
                                    )?                   # egroup, optional
                                    $                    # eol
                               """,
                               re.VERBOSE)


def read_base_line(line):
    if len(line.strip()) == 0:
        return None  # ignore empty lines
    if line[0] == b"#":
        return None  # ignore comments

    match = re.match(base_line_pattern, line)
    if match is None:
        raise MalformedCNF("Syntax error in line \"\"\"%s\"\"\"" % line)
    number, varname, instance, data, comment = match.groups()
    return {
        "number"   : marshal_in_number   (number),
        "varname"  : marshal_in_varname  (varname),
        "instance" : marshal_in_instance (instance),
        "data"     : marshal_in_data     (data),
        "comment"  : marshal_in_comment  (comment),
    }

child_line_pattern = re.compile(b"""
                                     ^                    # match from start
                                     \s*                  # optional spaces
                                     (\d+)                # line number
                                     \s+                  # spaces
                                     \((\d+)\)            # parent
                                     \s+                  # spaces
                                     ([A-Z][A-Z0-9_]*)    # varname
                                     \s*                  # optional spaces
                                     ,                    # delimiter
                                     \s*                  # optional spaces
                                     (-1|\d+)             # instance
                                     \s*                  # optional spaces
                                     :                    # delimiter
                                     \s*                  # optional spaces
                                     \"([^\"]*)\"         # quoted string (data)
                                     \s*                  # optional spaces
                                     (                    # bgroup
                                      \#                  # comment leader
                                      \s*                 # optional spaces
                                      .*                  # string (comment)
                                     )?                   # egroup, optional
                                     $                    # eol
                                """,
                                re.VERBOSE)


def read_child_line(line):
    if len(line.strip()) == 0:
        return None  # ignore empty lines
    if line[0] == "#":
        return None  # ignore comments

    match = re.match(child_line_pattern, line)
    if match is None:
        raise MalformedCNF("Syntax error in child line \"\"\"%s\"\"\""
                           % from_latin1 (line))
    number, parent, varname, instance, data, comment = match.groups()
    return {
        "number"   : marshal_in_number   (number),
        "parent"   : marshal_in_parent   (parent),
        "varname"  : marshal_in_varname  (varname),
        "instance" : marshal_in_instance (instance),
        "data"     : marshal_in_data     (data),
        "comment"  : marshal_in_comment  (comment),
    }


def parse_cnf_children(state, parent):
    lines = []
    current = get(state)
    while True:
        cnf_line = read_child_line(current)
        if cnf_line is not None:
            lines.append(cnf_line)
            state = advance(state)
            if state is None:
                break
            current = get(state)
            new_parent = get_parent(current)
            if new_parent is None:
                # drop stack
                return (state, lines, None)
            if new_parent > parent:
                # parent is further down in hierarchy -> new level
                (state, children, new_parent) = \
                    parse_cnf_children (state, new_parent)
                if state is None:
                    break
                cnf_line["children"] = children
                current = get(state)
                new_parent = get_parent(current)
                if new_parent is None:
                    # drop stack
                    return (state, lines, None)
            if new_parent < parent:
                # parent is further up in hierarchy -> pop level
                return (state, lines, new_parent)
            # new_parent == parent -> continue parsing on same level
    return (state, lines, parent)


def parse_cnf_root(state):
    lines = []
    current = get(state)
    while state:
        cnf_line = read_base_line(current)
        if cnf_line is not None:
            lines.append(cnf_line)
            state = advance(state)
            if state is None:  # -> nothing left to do
                break
            current = get(state)
            parent = get_parent(current)  # peek at next line
            if parent is not None:  # -> recurse into children
                (state, children, _parent) = parse_cnf_children(state, parent)
                cnf_line["children"] = children
                if state is None:
                    break
                current = get(state)
    return lines


def read_cnf(data):
    """
    Read cnf data from data bytes.

    :param bytes data: raw data
    :return: the parsed cnf data
    :rtype: {str, {str, str or int}}
    """
    if isinstance (data, str):
        data = to_latin1 (data)
    state = prepare(data)
    if state is None:
        raise InvalidCNF("Empty input string.")

    cnf = parse_cnf_root(state)
    if is_cnf(cnf) is False:
        raise TypeError("Invalid CNF_VAR.")
    return {"cnf": cnf}


def renumber_vars(root, parent=None, toplevel=False):
    """
    renumber_vars -- Number cnfvars linearly. If *parent* is specified,
    numbering will start at this offset. Also, the VAR *root* will be assigned
    this number as a parent lineno unless *toplevel* is set (the root var in
    a CNF tree obviously can’t have a parent).

    The *toplevel* parameter is useful when renumbering an existing variable
    starting at a given offset without at the same time having that offset
    assigned as a parent.
    """
    root = cnf_root (root)
    i = parent or 0
    for var in root:
        i += 1
        var["number"] = i
        if toplevel is False and parent is not None:
            var["parent"] = parent
        children = var.get("children", None)
        if children is not None:
            i = renumber_vars(children, parent=i, toplevel=False)
    return i

#
#                                serialization
#

cnf_line_nest_indent = "  "
cnf_line_base_fmt    = "%d %s,%d: \"%s\""
cnf_line_child_fmt   = "%d %s(%d) %s,%d: \"%s\""

def format_cnf_vars(da, var):
    """
    Return a list of formatted cnf_line byte strings.

    The variable names are uppercased unconditionally because while ``get_cnf``
    is case-indifferent for variable names, ``set_cnf`` isn’t.
    """

    depth, acc = da
    line = None
    if depth > 0:
        line = cnf_line_child_fmt \
            % (var["number"],
               cnf_line_nest_indent * depth,
               var["parent"],
               var["varname"].upper (),
               var["instance"],
               var["data"])
    else:
        line = cnf_line_base_fmt \
            % (var["number"],
               var["varname"].upper (),
               var["instance"],
               var["data"])

    comment = var.get("comment", None)
    if comment and len (comment) != 0:
        line = line + (" # %s" % comment)

    acc.append(to_latin1 (line))

    children = var.get("children", None)
    if children is not None:
        (_, acc) = functools.reduce(format_cnf_vars, children, (depth + 1, acc))

    return (depth, acc)


CNF_FIELD_MANDATORY = set ([ "varname", "data", "instance" ])
CNF_FIELD_OPTIONAL  = set ([ "parent", "children", "comment", "number" ])
CNF_FIELD_KNOWN     = CNF_FIELD_MANDATORY | CNF_FIELD_OPTIONAL


def is_cnf_var (obj):
    """
    Applies if all mandatory fields of a CNF_VAR and no unknown fields are
    present.

    .. XXX: validate field types.
    """
    assert isinstance (obj, dict)

    for f in CNF_FIELD_MANDATORY:
        if obj.get (f, None) is None:
            return False

    for f in obj:
        if f not in CNF_FIELD_KNOWN:
            return False

    return True


def cnf_root (root):
    """
    Of a given object, return the cnf root. This may be either a list
    containing the object itself, if it satisfies the criteria for a CNF_VAR,
    or the object contained in its toplevel field *cnf*.
    """
    if isinstance (root, list):
        return root
    if not isinstance(root, dict):
        raise TypeError(
            "Expected dictionary of CNF_VARs, got %s." % type(root))
    if is_cnf_var (root):
        return [ root ]
    cnf = root.get("cnf", None)
    if not isinstance(cnf, list):
        raise TypeError("Expected list of CNF_VARs, got %s." % type(cnf))
    return cnf


def normalize_cnf (cnf):
    """
    Ensure the output conforms to set_cnf()’s expectations.
    """
    if isinstance (cnf, list) is False:
        raise MalformedCNF ("expected list of CNF_VARs, got [%s]" % type (cnf))
    def norm (var):
        vvar = \
            { "number"   : var ["number"]
            , "varname"  : var ["varname"].upper ()
            , "instance" : var ["instance"]
            , "data"     : var ["data"]
            }

        children = var.get ("children", None)
        if children is not None:
            vvar ["children"] = normalize_cnf (children)

        return vvar

    return [ norm (var) for var in cnf ]


def print_cnf_raw(root, out=None):
    if root is not None:
        out.write(root)


def write_cnf_raw(*argv, **kw_argv):
    print_cnf_raw(*argv, **kw_argv)


def output_cnf(root, out, renumber=False, bytes=True):
    cnf = cnf_root(root)
    if renumber is True:
        _count = renumber_vars(root)
    if is_cnf(cnf) is True:
        (_, lines) = functools.reduce(format_cnf_vars, cnf, (0, []))
        if bytes is False:
            out.write ("\n".join (map (from_latin1, lines)))
            out.write ("\n")
        else:
            out.write (b"\n".join (lines))
            out.write (b"\n")


def dump_cnf_bytes (root, renumber=False):
    """
    dump_cnf_bytes -- Serialize CNF var structure, returning the result as a
    byte sequence.
    """
    cnf = cnf_root(root)
    out = io.BytesIO()
    output_cnf(root, out, renumber=renumber, bytes=True)
    res = out.getvalue()
    out.close()
    return res


def dump_cnf_string(root, renumber=False):
    """
    dump_cnf_string -- Serialize CNF var structure, returning the result as a
    string.
    """
    cnf = cnf_root(root)
    out = io.BytesIO()
    output_cnf(root, out, renumber=renumber, bytes=False)
    res = out.getvalue()
    out.close()
    return res

def print_cnf(root, out=None, renumber=False):
    if root is not None:
        output_cnf(root, out or sys.stdout, renumber=renumber, bytes=False)


def write_cnf(*argv, **kw_argv):
    print_cnf(*argv, **kw_argv)


def output_json(root, out, renumber=False):
    # Sanity check incompatible with Autotest. Who needs a seatbelt anyways?
    # if not isinstance(out, file):
        #raise TypeError("%s (%s) is not a stream." % (out, type(out)))
    if renumber is True:
        _count = renumber_vars(root)
    if is_cnf(root) is True:
        root ["cnf"] = normalize_cnf(cnf_root (root))
        data = json.dumps(root)
        out.write(data)
        out.write("\n")


def dump_json_string(root, renumber=False):
    """
    dump_json_string -- Serialize CNF var structure as JSON, returning the
    result as a string.
    """
    out = io.StringIO()
    output_json(root, out, renumber=renumber)
    res = out.getvalue()
    out.close()
    return res


def print_cnf_json(root, out=None, renumber=False):
    if root is not None:
        output_json(root, out or sys.stdout, renumber=renumber)


def write_cnf_json(*argv, **kw_argv):
    print_cnf_json(*argv, **kw_argv)

#
#                                  querying
#


def get_vars(cnf, data=None, instance=None):
    """
    get_vars -- Query a CNF_VAR structure. Skims the *toplevel* of the CNF_VAR
    structure for entries with a matching `data` or `instance` field.

    :type cnf: CNF_VAR
    :type data: str
    :type instance: int
    :rtype: CNF_VAR
    :returns: The structure containing only references to the
             matching variables. Containing an empty list of
             variables in case there is no match.
    """
    cnf = cnf["cnf"]
    if cnf:
        criterion = lambda _var: False
        if data:
            if instance:
                criterion = lambda var: var[
                    "data"] == data and var["instance"] == instance
            else:
                criterion = lambda var: var["data"] == data
        elif instance:
            criterion = lambda var: var["instance"] == instance

        return {"cnf": [var for var in cnf if criterion(var) is True]}

    return {"cnf": []}

def usage ():
    print ("usage: cnfvar.py -"      , file=sys.stderr)
    print (""                        , file=sys.stderr)
    print ("    Read CNF from stdin.", file=sys.stderr)
    print (""                        , file=sys.stderr)

#
#                         entry point for development
#


def main(argv):
    if len(argv) > 1:
        first = argv[1]
        if first == "-":
            cnf = read_cnf(sys.stdin.buffer.read())
            print_cnf(cnf)
            return 0
        elif first == "test":
            cnf = read_cnf(sys.stdin.buffer.read())
            cnff = get_vars(cnf, instance=2, data="FAX")
            print_cnf(cnff)
            return 0
    usage ()
    return -1

if __name__ == "__main__":
    sys.exit (main(sys.argv))
